﻿using System.Collections.Generic;
using UnityEngine;

namespace Oxide.Plugins
{
    [Info("Fly", "Mughisi", "1.2.1")]
    [Description("Allows players with permissions to fly around")]
    public class Fly : HurtworldPlugin
    {
        private readonly Hash<PlayerSession, PlayerInfo> players = new Hash<PlayerSession, PlayerInfo>();

        private const string permAllow = "fly.allowed";

#if HURTWORLDITEMV2
        private CharacterMotorConfiguration defaultConfig;
        private CharacterMotorConfiguration flyingConfig;
#endif

        private class PlayerInfo
        {
            public readonly PlayerSession Session;

            public bool IsFlying;
            public float BaseSpeed;

            public PlayerInfo(PlayerSession session)
            {
                Session = session;
                IsFlying = false;
                BaseSpeed = 75f;
            }
        }

        private void OnServerInitialized()
        {
            permission.RegisterPermission(permAllow, this);

            foreach (PlayerSession session in GameManager.Instance.GetSessions().Values)
            {
                players.Add(session, new PlayerInfo(session));
            }
        }

        private void Unload()
        {
            foreach (PlayerInfo player in players.Values)
            {
                if (player.IsFlying)
                {
                    SetFlymode(player.Session);
                }
            }
        }

        private void OnPlayerConnected(PlayerSession session)
        {
            if (!players.ContainsKey(session))
            {
                players.Add(session, new PlayerInfo(session));
            }
        }

        private void OnPlayerDisconnect(PlayerSession session)
        {
            if (players.ContainsKey(session))
            {
                players.Remove(session);
            }
        }

        private void OnPlayerInput(PlayerSession session, InputControls input)
        {
            if (!players.ContainsKey(session) || !players[session].IsFlying)
            {
                return;
            }

            CharacterMotorSimple motor = session.WorldPlayerEntity.GetComponent<CharacterMotorSimple>();
            EntityStats stats = session.WorldPlayerEntity.GetComponent<EntityStats>();

            if (!motor)
            {
                return;
            }

            Vector3 direction = new Vector3(IntFromBool(input.StrafeLeft) * -1 + IntFromBool(input.StrafeRight), 0f, IntFromBool(input.Backward) * -1 + IntFromBool(input.Forward));
            float speed = players[session].BaseSpeed;

            motor.IsGrounded = true;
#if HURTWORLDITEMV2
            direction = motor.RotationToAimDirectionCache * direction.normalized;
#else
            direction = motor.RotationToLookQuaternionXCache * direction.normalized;
#endif

            if (input.Forward)
            {
                direction.y = input.DirectionVector.y;
            }

            if (input.Backward)
            {
                direction.y = -input.DirectionVector.y;
            }

            if (input.Sprint)
            {
                speed *= 2;
            }

            if (motor._state.IsCrouching)
            {
                speed /= 2;
            }

            motor.Set_currentVelocity(motor.Accelerate(direction, speed, motor.GetVelocity(), 5, 5));

            if (!stats)
            {
                return;
            }

#if HURTWORLDITEMV2
            Dictionary<EntityFluidEffectKey, IEntityFluidEffect> effects = stats.GetFluidEffects();

            foreach (KeyValuePair<EntityFluidEffectKey, IEntityFluidEffect> effect in effects)
            {
                effect.Value.Reset(true);
            }
#else
            stats.GetFluidEffect(EEntityFluidEffectType.ColdBar).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.Radiation).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.HeatBar).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.Dampness).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.Hungerbar).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.InternalTemperature).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.ExternalTemperature).Reset(true);
            stats.GetFluidEffect(EEntityFluidEffectType.Toxin).Reset(true);
#endif
        }

        [ChatCommand("fly")]
        private void FlyCommand(PlayerSession session, string command, string[] args)
        {
            if (!session.IsAdmin && !permission.UserHasPermission(session.SteamId.ToString(), permAllow))
            {
                SendMessage(session, "No Permission");
                return;
            }

            if (args.Length > 0)
            {
                float speed;

                if (float.TryParse(args[0], out speed))
                {
                    SetFlymode(session, speed);
                    return;
                }
            }

            SetFlymode(session);
        }

        private Vector3 Ground(PlayerSession session)
        {
            CharacterMotorSimple motor = session.WorldPlayerEntity.GetComponent<CharacterMotorSimple>();
            Vector3 loc = motor.PlayerCamera.position;
            Vector3 dir = motor.PlayerCamera.TransformDirection(Vector3.down);
            RaycastHit hit;

            if (Physics.Raycast(loc, dir, out hit))
            {
                return new Vector3(hit.point.x, hit.point.y + 1f, hit.point.z);
            }

            return loc;
        }

        private void SetFlymode(PlayerSession session, float speed = 75f)
        {
            CharacterMotorSimple motor = session.WorldPlayerEntity.GetComponent<CharacterMotorSimple>();

            if (!motor)
            {
                return;
            }

#if HURTWORLDITEMV2
            if (defaultConfig == null && flyingConfig == null)
            {
                SetupMotorConfigs(motor);
            }
#endif

            players[session].IsFlying = !players[session].IsFlying;

            if (players[session].IsFlying)
            {
#if HURTWORLDITEMV2
                motor.Config = flyingConfig;
#else
                motor.GravityVector = Vector3.zero;
                motor.AirSpeedModifier = 1f;
                motor.FallDamageMultiplier = 0;
#endif
                players[session].BaseSpeed = speed;
            }
            else
            {
                session.WorldPlayerEntity.transform.position = Ground(session);
#if HURTWORLDITEMV2
                motor.Config = defaultConfig;
#else
                motor.GravityVector = new Vector3(0f, -25f, 0f);
                motor.AirSpeedModifier = 0.2f;
                motor.FallDamageMultiplier = 1.5f;
#endif
            }

            AlertManager.Instance.GenericTextNotificationServer(players[session].IsFlying
                    ? lang.GetMessage("Enabled", this, session.SteamId.ToString())
                    : lang.GetMessage("Disabled", this, session.SteamId.ToString()), session.Player);
        }

#if HURTWORLDITEMV2
        private void SetupMotorConfigs(CharacterMotorSimple motor)
        {
            defaultConfig = motor.Config;
            flyingConfig = UnityEngine.Object.Instantiate(defaultConfig);
            flyingConfig.AirSpeedModifier = 0.2f;
            flyingConfig.FallDamageMultiplier = 0f;
            flyingConfig.GravityVector = new Vector3(0f, -25f, 0f);
        }
#endif

        private new void LoadDefaultMessages()
        {
            // English
            lang.RegisterMessages(new Dictionary<string, string>
            {
                ["Enabled"] = "Fly mode enabled",
                ["Disabled"] = "Fly mode disabled",
                ["No Permission"] = "You don't have permission to use this command."
            }, this);

            // Polish
            lang.RegisterMessages(new Dictionary<string, string>
            {
                ["Enabled"] = "Włączono latanie",
                ["Disabled"] = "Wyłączono latanie",
                ["No Permission"] = "Nie posiadasz uprawnień do tej komendy."
            }, this, "pl");
        }

        private int IntFromBool(bool val) => !val ? 0 : 1;

        private void SendMessage(PlayerSession session, string message) => hurt.SendChatMessage(session, null, lang.GetMessage(message, this, session.SteamId.ToString()));
    }
}
